/* 
 *    Programmed By: Mohammed Isam Mohammed [mohammed_isam1984@yahoo.com]
 *    Copyright 2014, 2015 (c)
 * 
 *    file: calc.c
 *    This file is part of tcalc.
 *
 *    tcalc is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    tcalc is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with tcalc.  If not, see <http://www.gnu.org/licenses/>.
 */    

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>

extern int number_mode;
#ifndef __NUMBER_MODES
#define __NUMBER_MODES
#define NUMBER_INT	1
#define NUMBER_FLOAT	2
#endif /* __NUMBER_MODES */
extern double process_sub_formula_f(char* subformula, int end, char* error);
extern double process_math_func(char *func, int ob_index, int cb_index, char *error);

extern long hex_to_decimal(char *hex, char *error);
extern long binary_to_decimal(char *binary, char *error);
extern long octal_to_decimal(char *octal, char *error);
extern void decimal_to_hex(long decimal, char **res, char *error);
extern void decimal_to_binary(long decimal, char **res, char *error);


void do_op(char *tmp, char op, double *res, char *ERROR)
{
	double result_f = *res;
        switch(op)
	{
	  case '+':
	    if(tmp[0] == 'b' || tmp[0] == 'B')
	      result_f += binary_to_decimal(tmp+1, ERROR);
	    else if((tmp[0] == '0' && tmp[1] == 'X') ||
	      (tmp[0] == '0' && tmp[1] == 'x'))
	      result_f += hex_to_decimal(tmp, ERROR);
	    else if(tmp[0] == '0')
	      result_f += octal_to_decimal(tmp, ERROR);
	    else result_f += atof(tmp);
	    break;
	  case '-':
	    if(tmp[0] == 'b' || tmp[0] == 'B')
	      result_f -= binary_to_decimal(tmp+1, ERROR);
	    else if((tmp[0] == '0' && tmp[1] == 'X') ||
	      (tmp[0] == '0' && tmp[1] == 'x'))
	      result_f -= hex_to_decimal(tmp, ERROR);
	    else if(tmp[0] == '0')
	      result_f -= octal_to_decimal(tmp, ERROR);
	    else result_f -= atof(tmp);
	    break;
	  case '*':
	    if(tmp[0] == 'b' || tmp[0] == 'B')
	      result_f *= binary_to_decimal(tmp+1, ERROR);
	    else if((tmp[0] == '0' && tmp[1] == 'X') ||
	      (tmp[0] == '0' && tmp[1] == 'x'))
	      result_f *= hex_to_decimal(tmp, ERROR);
	    else if(tmp[0] == '0')
	      result_f *= octal_to_decimal(tmp, ERROR);
	    else result_f *= atof(tmp);
	    break;
	  case '/':
	    if(tmp[0] == 'b' || tmp[0] == 'B')
	      result_f /= binary_to_decimal(tmp+1, ERROR);
	    else if((tmp[0] == '0' && tmp[1] == 'X') ||
	      (tmp[0] == '0' && tmp[1] == 'x'))
	      result_f /= hex_to_decimal(tmp, ERROR);
	    else if(tmp[0] == '0')
	      result_f /= octal_to_decimal(tmp, ERROR);
	    else result_f /= atof(tmp);
	    break;
	  case '%':
	    if(tmp[0] == 'b' || tmp[0] == 'B')
	      result_f = (int)result_f%binary_to_decimal(tmp+1, ERROR);
	    else if((tmp[0] == '0' && tmp[1] == 'X') ||
	      (tmp[0] == '0' && tmp[1] == 'x'))
	      result_f = (int)result_f%hex_to_decimal(tmp, ERROR);
	    else if(tmp[0] == '0')
	      result_f = (int)result_f%octal_to_decimal(tmp, ERROR);
	    else result_f = (int)result_f%atoi(tmp);
	    break;
	  case '.':
	    //default:
	    if(tmp[0] == 'b' || tmp[0] == 'B')
	      result_f = binary_to_decimal(tmp+1, ERROR);
	    else if((tmp[0] == '0' && tmp[1] == 'X') ||
	      (tmp[0] == '0' && tmp[1] == 'x'))
	      result_f = hex_to_decimal(tmp, ERROR);
	    else if(tmp[0] == '0')
	      result_f = octal_to_decimal(tmp, ERROR);
	    else result_f = atof(tmp);
	    break;
	}
	*res = result_f;
}


double calc(char *formula, char *ERROR, char *COMMENT)
{
  int i = 0, k;
  char *tmp;
  long result = 0;
  double result_f = 0;
  char op = '.';
  number_mode = 0;
  *ERROR = 0;
  int FORMULA_LEN = strlen(formula);
  char special_number = 0;	/* flag to denote if we are in the middle of binary or hex num */
  for(k = 0; k <= strlen(formula); k++)
  {
    if(formula[k] == '(')
    {
      if(!strrchr(formula, ')'))
      {
	fprintf(stderr, "Error: Missing ')'\n");
	result = -1; result_f = -1; *ERROR = 1; break;
      }
      int j;
      int open_braces = 1, close_braces = 0;
      for(j = k+1; j < strlen(formula); j++)
      {
	if(formula[j] == ')') close_braces++;
	else if(formula[j] == '(') open_braces++;
	if(open_braces == close_braces) break;
      }
      double tmp_res = 
	process_sub_formula_f(formula+k, j-k, ERROR);
      if(*ERROR) break;
      if(strchr(formula+k, '.'))
	if(number_mode == NUMBER_INT)
	{ result_f = result; number_mode = NUMBER_FLOAT; }

      if(number_mode == NUMBER_FLOAT)
      {
	switch(op)
	{
	  case '+': result_f += tmp_res; break;
	  case '-': result_f -= tmp_res; break;
	  case '*': result_f *= tmp_res; break;
	  case '/': result_f /= tmp_res; break;
	  case '%': result_f = (int)result_f%(int)tmp_res; break;
	  case '.':
	    //default: 
	    result_f = tmp_res; break;
	}
      }
      else
      {
	switch(op)
	{
	  case '+': result += tmp_res; break;
	  case '-': result -= tmp_res; break;
	  case '*': result *= tmp_res; break;
	  case '/':
	    result_f = result; number_mode = NUMBER_FLOAT;
	    result_f /= tmp_res;
	    break;
	  case '%': result = result%(int)tmp_res; break;
	  case '.':
	    //default: 
	    result = tmp_res; break;
	}
      }
      k = j;//strrchr(formula+k, ')')-(formula);
      op = formula[k];
      i = k+1;
      if(special_number) special_number = 0;
    }
    /////////////////////////////////
    /////////////////////////////////
    /////////////////////////////////
    else if(formula[k] == '+' || formula[k] == '-' ||
	    formula[k] == '*' || formula[k] == '/' ||
	    formula[k] == '%' || formula[k] == '\0')
    {
      tmp = strndup(formula+i, k-i);
      if(number_mode == NUMBER_FLOAT)
      {
calc_float:
        do_op(tmp, op, &result_f, ERROR);
      }
      else
      {
	if(strchr(tmp, '.'))
	{
	  number_mode = NUMBER_FLOAT;
	  result_f = result; result = 0;
	  goto calc_float;
	}
	else
	{
	  number_mode = NUMBER_INT;
	  result_f = result;
          do_op(tmp, op, &result_f, ERROR);
	  result = (int)result_f;
	  if(op == '/')
	    number_mode = NUMBER_FLOAT;
	}
      }
      if(*ERROR) break;
      op = formula[k];
      i = k+1;
      if(special_number) special_number = 0;
    }
    ///////////////////////////////////////
    //process math functions
    ///////////////////////////////////////
    else if(formula[k] >= 'a' && formula[k] <= 'z')
    {
      /* Maybe we are misreading a binary or decimal number? */
      if(special_number) continue;
      if(k < FORMULA_LEN-1 && formula[k] == 'b'&&
	(formula[k+1] == '1' || formula[k+1] == '0')) { special_number = 1; continue; }
      if(k && formula[k-1] == '0' && (formula[k] == 'x' || formula[k] == 'X')) { special_number = 1; continue; }
      /* Maybe this is a math function name? */
      char *ob = strchr(formula+k, '(');
      char *cb = strchr(formula+k, ')');
      if(!ob || !cb)
      {
	/* Nope. The user messed up. Tell him! */
	if(!ob) printf("Error: missing '('\n");
	else    printf("Error: missing ')'\n");
	result = -1; result_f = -1; *ERROR = 1; break;
      }
      double r =
	process_math_func(formula+k, ob-(formula+k), cb-(formula+k), ERROR);
      if(*ERROR) break;
      if(number_mode == NUMBER_INT)
      { result_f = result; }
      number_mode = NUMBER_FLOAT;
      switch(op)
      {
	case '+': result_f += r; break;
	case '-': result_f -= r; break;
	case '*': result_f *= r; break;
	case '/': result_f /= r; break;
	case '%': result_f = (int)result_f%(int)r; break;
	case '.':
	  //default: 
	  result_f = r; break;
      }
      k = cb-(formula+k)+1;
      op = formula[k];
      i = k+1;
    }
    else if(formula[k] == ' ' || formula[k] == '\t') i++;
    //else if(formula[k] == '#') { COMMENT = 1; break; }
  }//end for
  
  if(number_mode == NUMBER_INT)
    result_f = result;
  return result_f;
}