/* 
 *    Programmed By: Mohammed Isam Mohammed [mohammed_isam1984@yahoo.com]
 *    Copyright 2014, 2015 (c)
 * 
 *    file: main.c
 *    This file is part of tcalc.
 *
 *    tcalc is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    tcalc is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with tcalc.  If not, see <http://www.gnu.org/licenses/>.
 */    

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <getopt.h>
#include <asm/types.h>
#include <limits.h>
#include <string.h>
#include <math.h>

const char tcalc_ver[] = "1.5";
int input_method;
#define INPUT_FROM_FILE	1
#define INPUT_FORMULA	2
#define INPUT_FREE	3

int number_mode = 0;
#ifndef __NUMBER_MODES
#define __NUMBER_MODES
#define NUMBER_INT	1
#define NUMBER_FLOAT	2
#endif /* __NUMBER_MODES */

int decimal_point = 0;
#ifndef __FORMATS
#define __FORMATS
#define DECIMAL_FORMAT	1
#define HEX_FORMAT	2
#define BINARY_FORMAT	3
#define OCTAL_FORMAT	4
#define ALL_FORMATS	5
#endif /* __FORMATS */
int input_format = DECIMAL_FORMAT;
int output_format = DECIMAL_FORMAT;

FILE *file;

double process_sub_formula_f(char* subformula, int end, char* error);
void print_mult_table(char *n);
extern void input_free_mode();
extern void input_from_file_mode(FILE *file);

extern long hex_to_decimal(char *hex, char *error);
extern long binary_to_decimal(char *binary, char *error);
extern long octal_to_decimal(char *octal, char *error);
extern void decimal_to_hex(long decimal, char **res, char *error);
extern void decimal_to_binary(long decimal, char **res, char *error);
//extern long decimal_to_octal(long decimal);
extern void convert_numbers(char *input);
extern double process_math_func(char *func, int ob_index, int cb_index, char *error);
extern void print_func_descriptions();

//for custom functions
#ifndef __CUSTOM_FUNCS
#define __CUSTOM_FUNCS
#define MAX_FUNCTIONS		20
#define MAX_FUNCTION_NAME	10
#define MAX_VARIABLES		10
#define MAX_VARIABLE_LEN	10
#endif /* __CUSTOM_FUNCS */

extern char custom_func_name[][MAX_FUNCTION_NAME];
extern char *func_body[];
extern char *func_args[];
extern int func_args_count[];
extern int func_body_len[];
extern char var_name[][MAX_VARIABLE_LEN];
extern float var_value[];
extern int total_vars;
extern int total_custom_funcs;
extern int process_custom_func(char *name, char *str, char *args, double *res);
extern double calc(char *formula, char *ERROR, char *COMMENT);		/* calc.c */
extern void do_op(char *tmp, char op, double *res, char *ERROR);	/* calc.c */

//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
int main(int argc, char **argv)
{
	///////////////////////////////////////
	//parse command line arguments
	///////////////////////////////////////
	int c;
	char *file_name = NULL;
	char *formula = NULL;
	static int READ_FILE = 0, READ_FORMULA = 0, QUIET = 0;
	static int CONVERT = 0;
	char *num_to_convert = NULL;
	static int MULT_TABLE = 0;
	char *mult_num = NULL;
	static struct option long_options[] =
	{
	 {"input",   required_argument, &READ_FILE,  1},
	 {"version", no_argument,       0,         'v'},
	 {"help",    no_argument,       0,         'h'},
	 {"quiet",   no_argument,       0,         'q'},
	 {"decimal", required_argument, 0,         'd'},
	 {"format",  required_argument, 0,         'f'},
	 {"output",  required_argument, 0,         'o'},
	 {"convert", required_argument, 0,         'c'},
	 {"math",    no_argument,       0,         'm'},
	 {"table",   required_argument, 0,         't'},
	 {0, 0, 0, 0}
	};

	while(1)
	{
		int option_index=0;
		c = getopt_long(argc, argv, "i:vhqd:f:o:c:mt:", long_options, &option_index);
		if(c==-1) break;	//end of options

		switch(c)
		{
		case 0:
			if(long_options[option_index].flag != 0)
				if(strcmp(long_options[option_index].name, "input") == 0)
					file_name = optarg;
			break;
		case 'i':	//save file name so we can display info later
			file_name = optarg;
			READ_FILE = 1; break;
		case 'd':	//define decimal places
			decimal_point = atoi(optarg);
			break;
		case 't':	//print multiplication table
			mult_num = optarg;
			MULT_TABLE = 1;
			break;
		case 'c':	//convert a number
			CONVERT = 1;
			num_to_convert = optarg;
			break;
		case 'f':	//define input format
			if(strcmp(optarg, "d") == 0 || strcmp(optarg, "decimal") == 0)
			  input_format = DECIMAL_FORMAT;
			else if(strcmp(optarg, "x") == 0 || strcmp(optarg, "hex") == 0)
			  input_format = HEX_FORMAT;
			else if(strcmp(optarg, "b") == 0 || strcmp(optarg, "binary") == 0)
			  input_format = BINARY_FORMAT;
			else if(strcmp(optarg, "o") == 0 || strcmp(optarg, "octal") == 0)
			  input_format = OCTAL_FORMAT;
			else
			{
			  printf("Error: Unknown number format\n");
			  exit(1);
			}
			break;
		case 'o':	//define output format
			if(strcmp(optarg, "d") == 0 || strcmp(optarg, "decimal") == 0)
			  output_format = DECIMAL_FORMAT;
			else if(strcmp(optarg, "x") == 0 || strcmp(optarg, "hex") == 0)
			  output_format = HEX_FORMAT;
			else if(strcmp(optarg, "b") == 0 || strcmp(optarg, "binary") == 0)
			  output_format = BINARY_FORMAT;
			else if(strcmp(optarg, "o") == 0 || strcmp(optarg, "octal") == 0)
			  output_format = OCTAL_FORMAT;
			else if(strcmp(optarg, "a") == 0 || strcmp(optarg, "all") == 0)
			  output_format = ALL_FORMATS;
			else
			{
			  printf("Error: Unknown number format\n");
			  exit(1);
			}
			break;
		case 'v':	//show version & exit
			printf("Version %s\n", tcalc_ver);
			exit(0); break;
		case 'm':	//print math functions & exit
			print_func_descriptions();
			exit(0); break;
		case 'q':	//quiet output
			QUIET = 1;
			break;
		case 'h':	//show help & exit
			printf("Terminal Calculator for GNU/Linux console, version %s\n"
				"Usage: %s [options]\n\n"
				"Options:\n"
				"Formula              Formula to be calculated\n"
				"-c, --convert n      Convert n to the output format specified\n"
				"                     by '-f' below, or to decimal if none\n"
				"-d, --decimal n      How many digits are shown to the right\n"
				"                     of the decimal point\n"
				"-f, --format n       Input format, possible values are:\n"
				"                           * binary (b)\n"
				"                           * hex (x)\n"
				"                           * decimal (d)\n"
				"                           * octal (o)\n"
				"-h, --help           Print help [this page]\n"
				"-i, --input filename Name of file containing formulas\n"
				"-m, --math           Print a list of math functions\n"
				"-o, --output n       Output format, possible values are:\n"
				"                           * binary (b)\n"
				"                           * hex (x)\n"
				"                           * decimal (d)\n"
				"                           * octal (o)\n"
				"                           * all formats (a)\n"
				"-q, --quiet          Minimize output messages\n"
				"-t, --table n        Print the multiplication table of number 'n'\n"
				"-v, --version        Display program version\n",
				tcalc_ver, argv[0]);
			exit(0); break;
		case '?':
			break;
		default:
			abort();
		}
	}
	///////////////////////////////////////
	while(optind < argc)
	{
	  formula = argv[optind++];
	  READ_FORMULA = 1;
	}

	if(file_name == NULL)
	{
	  if(!READ_FORMULA)
	    input_method = INPUT_FREE;
	  else
	    input_method = INPUT_FORMULA;
	}
	else
	{
	  if((file = fopen(file_name, "rb")) == NULL)
	  {
	    fprintf(stderr, "Error opening file: %s\n", file_name);
	    exit(1);
	  }
	  else
	    input_method = INPUT_FROM_FILE;
	}
	
	/*if(strchr(formula, '\n'))
	{
	  printf("formula=%s\n", formula);
	  input_method = INPUT_FROM_FILE;
	  file = stdin;
	}*/

	if(CONVERT)
	{
	  //printf("n=%s, input=%d\n", num_to_convert, input_format);
	  convert_numbers(num_to_convert);
	  exit(0);
	}
	
	if(MULT_TABLE)
	{
	  print_mult_table(mult_num);
	  exit(0);
	}
	
	if(!QUIET)
	{
	  printf("Terminal Calculator for GNU/Linux console\n");
	  if(input_method == INPUT_FROM_FILE)
	    printf("Reading from %s\n", file_name);
	  else if(input_method == INPUT_FORMULA)
	    printf("Reading command line formula\n");
	  else
	    printf("Free input:\nEnter 'q' to quit, 'h' for help\n");
	}
	
	///////////////////////////////////////
	///////////////////////////////////////
	if(input_method == INPUT_FROM_FILE)
	{
	  input_from_file_mode(file);
	  fclose(file);
	}
	else if(input_method == INPUT_FREE)
	{
	  input_free_mode();
	}
	else if(input_method == INPUT_FORMULA)
	{
	  char COMMENT = 0;
	  char error = 0;
	  long result = 0;
	  double result_f = 0;
	  result_f = calc(formula, &error, &COMMENT);
	  if(number_mode == NUMBER_INT)
	    result = (int)result_f;
	/*  
	  int i = 0, k;
	  char *tmp;
	  long result = 0;
	  double result_f = 0;
	  char op = '.';
	  number_mode = 0;
	  char error;
	  char special_number = 0;	// flag to denote if we are in the middle of binary or hex num //
	  int FORMULA_LEN = strlen(formula);
	  //while(i < strlen(formula))
	  //{
	    for(k = 0; k <= strlen(formula); k++)
	    {
	      if(formula[k] == '(')
	      {
		if(!strrchr(formula, ')'))
		{
		  fprintf(stderr, "Error: Missing ')'\n");
		  exit(1);
		}
		int j;
		int open_braces = 1, close_braces = 0;
		for(j = k+1; j < strlen(formula); j++)
		{
		  if(formula[j] == ')') close_braces++;
		  else if(formula[j] == '(') open_braces++;
		  if(open_braces == close_braces) break;
		}
		double tmp_res = 
		  process_sub_formula_f(formula+k, j-k, &error);
		if(error) exit(1);
		  //process_sub_formula_f(formula+k, strrchr(formula+k, ')')-(formula+k));
		if(strchr(formula+k, '.'))
		  if(number_mode == NUMBER_INT)
		  { result_f = result; number_mode = NUMBER_FLOAT; }
		
		if(number_mode == NUMBER_FLOAT)
		{
		  switch(op)
		  {
		    case '+': result_f += tmp_res; break;
		    case '-': result_f -= tmp_res; break;
		    case '*': result_f *= tmp_res; break;
		    case '/': result_f /= tmp_res; break;
		    case '%': result_f = (int)result_f%(int)tmp_res; break;
		    case '.': result_f = tmp_res; break;
		  }
		}
		else
		{
		  switch(op)
		  {
		    case '+': result += tmp_res; break;
		    case '-': result -= tmp_res; break;
		    case '*': result *= tmp_res; break;
		    case '/': result /= tmp_res; break;
		    case '%': result = result%(int)tmp_res; break;
		    case '.': result = tmp_res; break;
		  }
		}
		k = j;//strrchr(formula+k, ')')-(formula);
		op = formula[k];
		i = k+1;
	      }
	      /////////////////////////////////
	      /////////////////////////////////
	      /////////////////////////////////
	      else if(formula[k] == '+' || formula[k] == '-' ||
		      formula[k] == '*' || formula[k] == '/' ||
		      formula[k] == '%' || formula[k] == '\0')
	      {
		tmp = strndup(formula+i, k-i);
		if(number_mode == NUMBER_FLOAT)
		{
calc_float:
		  switch(op)
		  {
		    case '+':
		      if(tmp[0] == 'b' || tmp[0] == 'B')
			result_f += binary_to_decimal(tmp+1, &error);
		      else if((tmp[0] == '0' && tmp[1] == 'X') ||
			      (tmp[0] == '0' && tmp[1] == 'x'))
			result_f += hex_to_decimal(tmp, &error);
		      else if(tmp[0] == '0')
			result_f += octal_to_decimal(tmp, &error);
		      else result_f += atof(tmp);
		      break;
		    case '-':
		      if(tmp[0] == 'b' || tmp[0] == 'B')
			result_f -= binary_to_decimal(tmp+1, &error);
		      else if((tmp[0] == '0' && tmp[1] == 'X') ||
			      (tmp[0] == '0' && tmp[1] == 'x'))
			result_f -= hex_to_decimal(tmp, &error);
		      else if(tmp[0] == '0')
			result_f -= octal_to_decimal(tmp, &error);
		      else result_f -= atof(tmp);
		      break;
		    case '*':
		      if(tmp[0] == 'b' || tmp[0] == 'B')
			result_f *= binary_to_decimal(tmp+1, &error);
		      else if((tmp[0] == '0' && tmp[1] == 'X') ||
			      (tmp[0] == '0' && tmp[1] == 'x'))
			result_f *= hex_to_decimal(tmp, &error);
		      else if(tmp[0] == '0')
			result_f *= octal_to_decimal(tmp, &error);
		      else result_f *= atof(tmp);
		      break;
		    case '/':
		      if(tmp[0] == 'b' || tmp[0] == 'B')
			result_f /= binary_to_decimal(tmp+1, &error);
		      else if((tmp[0] == '0' && tmp[1] == 'X') ||
			      (tmp[0] == '0' && tmp[1] == 'x'))
			result_f /= hex_to_decimal(tmp, &error);
		      else if(tmp[0] == '0')
			result_f /= octal_to_decimal(tmp, &error);
		      else result_f /= atof(tmp);
		      break;
		    case '%':
		      if(tmp[0] == 'b' || tmp[0] == 'B')
			result_f = (int)result_f%binary_to_decimal(tmp+1, &error);
		      else if((tmp[0] == '0' && tmp[1] == 'X') ||
			      (tmp[0] == '0' && tmp[1] == 'x'))
			result_f = (int)result_f%hex_to_decimal(tmp, &error);
		      else if(tmp[0] == '0')
			result_f = (int)result_f%octal_to_decimal(tmp, &error);
		      else result_f = (int)result_f%atoi(tmp);
		      break;
		    case '.':
		      if(tmp[0] == 'b' || tmp[0] == 'B')
			result_f = binary_to_decimal(tmp+1, &error);
		      else if((tmp[0] == '0' && tmp[1] == 'X') ||
			      (tmp[0] == '0' && tmp[1] == 'x'))
			result_f = hex_to_decimal(tmp, &error);
		      else if(tmp[0] == '0')
			result_f = octal_to_decimal(tmp, &error);
		      else result_f = atof(tmp);
		      break;
		  }
		}
		else
		{
		  if(strchr(tmp, '.'))
		  {
		    number_mode = NUMBER_FLOAT;
		    result_f = result; result = 0;
		    goto calc_float;
		  }
		  else
		  {
		    number_mode = NUMBER_INT;
		    switch(op)
		    {
		      case '+':
			if(tmp[0] == 'b' || tmp[0] == 'B')
			  result += binary_to_decimal(tmp+1, &error);
			else if((tmp[0] == '0' && tmp[1] == 'X') ||
				(tmp[0] == '0' && tmp[1] == 'x'))
			  result += hex_to_decimal(tmp, &error);
			else if(tmp[0] == '0')
			  result += octal_to_decimal(tmp, &error);
			else result += atoi(tmp);
			break;
		      case '-':
			if(tmp[0] == 'b' || tmp[0] == 'B')
			  result -= binary_to_decimal(tmp+1, &error);
			else if((tmp[0] == '0' && tmp[1] == 'X') ||
				(tmp[0] == '0' && tmp[1] == 'x'))
			  result -= hex_to_decimal(tmp, &error);
			else if(tmp[0] == '0')
			  result -= octal_to_decimal(tmp, &error);
			else result -= atoi(tmp);
			break;
		      case '*':
			if(tmp[0] == 'b' || tmp[0] == 'B')
			  result *= binary_to_decimal(tmp+1, &error);
			else if((tmp[0] == '0' && tmp[1] == 'X') ||
				(tmp[0] == '0' && tmp[1] == 'x'))
			  result *= hex_to_decimal(tmp, &error);
			else if(tmp[0] == '0')
			  result *= octal_to_decimal(tmp, &error);
			else result *= atoi(tmp);
			break;
		      case '/':
			if(tmp[0] == 'b' || tmp[0] == 'B')
			  result /= binary_to_decimal(tmp+1, &error);
			else if((tmp[0] == '0' && tmp[1] == 'X') ||
				(tmp[0] == '0' && tmp[1] == 'x'))
			  result /= hex_to_decimal(tmp, &error);
			else if(tmp[0] == '0')
			  result /= octal_to_decimal(tmp, &error);
			else result /= atoi(tmp);
			break;
		      case '%':
			if(tmp[0] == 'b' || tmp[0] == 'B')
			  result = (int)result%binary_to_decimal(tmp+1, &error);
			else if((tmp[0] == '0' && tmp[1] == 'X') ||
				(tmp[0] == '0' && tmp[1] == 'x'))
			  result = (int)result%hex_to_decimal(tmp, &error);
			else if(tmp[0] == '0')
			  result = (int)result%octal_to_decimal(tmp, &error);
			else result = (int)result%atoi(tmp);
			break;
		      case '.':
			if(tmp[0] == 'b' || tmp[0] == 'B')
			  result = binary_to_decimal(tmp+1, &error);
			else if((tmp[0] == '0' && tmp[1] == 'X') ||
				(tmp[0] == '0' && tmp[1] == 'x'))
			  result = hex_to_decimal(tmp, &error);
			else if(tmp[0] == '0')
			  result = octal_to_decimal(tmp, &error);
			else result = atoi(tmp);
			break;
		    }
		  }
		}
		if(error) break;
		op = formula[k];
		i = k+1;
		if(special_number) special_number = 0;
	      }
	      ///////////////////////////////////////
	      //process math functions
	      ///////////////////////////////////////
	      else if(formula[k] >= 'a' && formula[k] <= 'z')
	      {
		// Maybe we are misreading a binary or decimal number? //
		if(special_number) continue;
		if(k < FORMULA_LEN-1 && formula[k] == 'b'&&
		  (formula[k+1] == '1' || formula[k+1] == '0')) { special_number = 1; continue; }
		if(k && formula[k-1] == '0' && (formula[k] == 'x' || formula[k] == 'X')) { special_number = 1; continue; }
		// Maybe this is a math function name? //
		char *ob = strchr(formula+k, '(');
		char *cb = strchr(formula+k, ')');
		if(!ob || !cb)
		{
		  // Nope. The user messed up. Tell him! //
		  if(!ob) printf("Error: missing '('\n");
		  else    printf("Error: missing ')'\n");
		  exit(1);
		}
		//char error;
		double r =
		  process_math_func(formula+k, ob-(formula+k), cb-(formula+k), &error);
		if(error) exit(1);
		if(number_mode == NUMBER_INT)
		{ result_f = result; }
		number_mode = NUMBER_FLOAT;
		switch(op)
		{
		    case '+': result_f += r; break;
		    case '-': result_f -= r; break;
		    case '*': result_f *= r; break;
		    case '/': result_f /= r; break;
		    case '%': result_f = (int)result_f%(int)r; break;
		    case '.': result_f = r; break;
		}
		k = cb-(formula+k)+1;
		op = formula[k];
		i = k+1;
	      }
	    }//end for
	    */
	  
	  //////////////////////////////////////////////
	  //output the result
	  //////////////////////////////////////////////
	  if(error) exit(1);
	  
	  if(output_format == DECIMAL_FORMAT || output_format == ALL_FORMATS)
	  {
	    if(decimal_point)
	    {
	      char tmp2[20], tmp3[10];
	      tmp2[0] = '%';
	      tmp2[1] = '.';
	      tmp2[2] = '\0';
	      sprintf(tmp3, "%d", decimal_point);
	      strcat(tmp2, tmp3);
	      strcat(tmp2, "f\n");
	      //printf("%s %s\n", tmp3, tmp2);
	      if(number_mode == NUMBER_INT)
		result_f = result;
	      printf(tmp2, result_f);
	    }
	    else
	    {
	      if(number_mode == NUMBER_INT) printf("%ld\n", result);
	      else printf("%f\n", result_f);
	    }
	  }
	  if(output_format == BINARY_FORMAT || output_format == ALL_FORMATS)
	  {
	      char *res2;
	      if(number_mode == NUMBER_INT)
	      {
		decimal_to_binary(result, &res2, &error);
		if(res2) { printf("b%s\n", res2); free(res2); }
	      }
	      else
	      {
		decimal_to_binary((long)result_f, &res2, &error);
		if(res2) { printf("b%s\n", res2); free(res2); }	
	      }
	  }
	  if(output_format == OCTAL_FORMAT || output_format == ALL_FORMATS)
	  {
	    if(number_mode == NUMBER_INT)
	      printf("0%o\n", (unsigned)result);
	    else printf("0%o\n", (unsigned int)result_f);
	  }
	  if(output_format == HEX_FORMAT || output_format == ALL_FORMATS)
	  {
	      char *res2;
	      if(number_mode == NUMBER_INT)
	      {
		decimal_to_hex(result, &res2, &error);
		if(res2) { printf("%s\n", res2); free(res2); }
	      }
	      else
	      {
		decimal_to_hex((long)result_f, &res2, &error);
		if(res2) { printf("%s\n", res2); free(res2); }	
	      }
	  }
	}//end INPUT_FORMULA
	
	return 0;
}


///////////////////////////////////////////
///////////////////////////////////////////
///////////////////////////////////////////
double process_sub_formula_f(char *subformula, int end, char *error)
{
  int i = 1, j, k;
  double res = 0;
  char op = '.';
  char *tmp;
  *error = 0;
  for(k = 1; k <= end; k++)
  {
      if(subformula[k] == '(')
      {
	for(j = end-1; j > 0; j--)
	  if(subformula[j] == ')')
	  {
		double tmp_res = 
		  process_sub_formula_f(subformula+k, j-k, error);
		if(*error) return 0;
		
		switch(op)
		{
		    case '+': res += tmp_res; break;
		    case '-': res -= tmp_res; break;
		    case '*': res *= tmp_res; break;
		    case '/': res /= tmp_res; break;
		    case '%': res = (int)res%(int)tmp_res; break;
		    case '.': res = tmp_res; break;
		}
		break;
	  }
	  if(j <= 0)
	  {
		  fprintf(stderr, "Error: Missing ')'\n");
		  *error = 1;
		  return 0;
	  }
	  op = subformula[j+1];
	  k = j+1;
	  i = j+2;
      }
      /////////////////////////////////
      /////////////////////////////////
      /////////////////////////////////
      else if(subformula[k] == '+' || subformula[k] == '-' ||
	      subformula[k] == '*' || subformula[k] == '/' ||
	      subformula[k] == '%' || subformula[k] == '\0'||
	      subformula[k] == ')')
      {
	tmp = strndup(subformula+i, k-i);
	if(!tmp) { *error = 2; return 0; }
	
	//check if this is a variable
	if(tmp[0] == 'a' || (tmp[0] > 'b' && tmp[0] <= 'z'))
	{
	  int m;
	  for(m = 0; m < total_vars; m++)
	  {
	    if(strcmp(var_name[m], tmp) == 0)
	    {
	      sprintf(tmp, "%f", var_value[m]); break;
	    }
	    //int n;
	    //for(n = 0; n < strlen(var_name[m]); n++)
	      //printf("%s, %d\n", var_name[m], strncmp(var_name[m], tmp, strlen(var_name[m])));
	  }
	  if(m == total_vars)
	  {
	    printf("Error: variable %s is not defined, %d\n", tmp, total_vars);
	    *error = 3;
	    return 0;
	  }
	}
	
	do_op(tmp, op, &res, error);
	/*
	switch(op)
	{
	      case '+':
		if(tmp[0] == 'b' || tmp[0] == 'B')
		  res += binary_to_decimal(tmp+1, error);
		else if((tmp[0] == '0' && tmp[1] == 'X') ||
			(tmp[0] == '0' && tmp[1] == 'x'))
		  res += hex_to_decimal(tmp, error);
		else if(tmp[0] == '0')
		  res += octal_to_decimal(tmp, error);
		else res += atof(tmp);
		break;
	      case '-':
		if(tmp[0] == 'b' || tmp[0] == 'B')
		  res -= binary_to_decimal(tmp+1, error);
		else if((tmp[0] == '0' && tmp[1] == 'X') ||
			(tmp[0] == '0' && tmp[1] == 'x'))
		  res -= hex_to_decimal(tmp, error);
		else if(tmp[0] == '0')
		  res -= octal_to_decimal(tmp, error);
		else res -= atof(tmp);
		break;
	      case '*':
		if(tmp[0] == 'b' || tmp[0] == 'B')
		  res *= binary_to_decimal(tmp+1, error);
		else if((tmp[0] == '0' && tmp[1] == 'X') ||
			(tmp[0] == '0' && tmp[1] == 'x'))
		  res *= hex_to_decimal(tmp, error);
		else if(tmp[0] == '0')
		  res *= octal_to_decimal(tmp, error);
		else res *= atof(tmp);
		break;
	      case '/':
		if(tmp[0] == 'b' || tmp[0] == 'B')
		  res /= binary_to_decimal(tmp+1, error);
		else if((tmp[0] == '0' && tmp[1] == 'X') ||
			(tmp[0] == '0' && tmp[1] == 'x'))
		  res /= hex_to_decimal(tmp, error);
		else if(tmp[0] == '0')
		  res /= octal_to_decimal(tmp, error);
		else res /= atof(tmp);
		break;
	      case '%':
		if(tmp[0] == 'b' || tmp[0] == 'B')
		  res = (int)res%binary_to_decimal(tmp+1, error);
		else if((tmp[0] == '0' && tmp[1] == 'X') ||
			(tmp[0] == '0' && tmp[1] == 'x'))
		  res = (int)res%hex_to_decimal(tmp, error);
		else if(tmp[0] == '0')
		  res = (int)res%octal_to_decimal(tmp, error);
		else res = (int)res%atoi(tmp);
		break;
	      case '.':
		if(tmp[0] == 'b' || tmp[0] == 'B')
		  res = binary_to_decimal(tmp+1, error);
		else if((tmp[0] == '0' && tmp[1] == 'X') ||
			(tmp[0] == '0' && tmp[1] == 'x'))
		  res = hex_to_decimal(tmp, error);
		else if(tmp[0] == '0')
		  res = octal_to_decimal(tmp, error);
		else res = atof(tmp);
		break;
	}
	*/
	op = subformula[k];
	i = k+1;
      }
  }//end for
  printf("process_sub_formula_f(): res = %f\n", res);
  return res;
}


void print_mult_table(char *n)
{
  //printf("In print_mult_table(), n = '%s'\n", n);
  long num;
  char ERROR;
  if(n[0] == 'b' || n[0] == 'B')
    num = binary_to_decimal(n+1, &ERROR);
  else if((n[0] == '0' && n[1] == 'X') ||
	  (n[0] == '0' && n[1] == 'x'))
    num = hex_to_decimal(n, &ERROR);
  else if(n[0] == '0')
    num = octal_to_decimal(n, &ERROR);
  else num = atol(n);
  
  //printf("%ld, %d\n", num, number_mode);
  int i;
  char *r, *r2, *r3;
  for(i = 1; i <= 12; i++)
  {
    if(output_format == DECIMAL_FORMAT)
      printf("%ld * %2d = %ld\n", num, i, num*i);
    else if(output_format == BINARY_FORMAT)
    {
      decimal_to_binary(num,   &r, &ERROR);
      decimal_to_binary(i,     &r2, &ERROR);
      decimal_to_binary(num*i, &r3, &ERROR);
      printf("b%s * b%s = b%s\n", r, r2, r3);
      free(r); free(r2); free(r3);
    }
    else if(output_format == HEX_FORMAT)
    {
      decimal_to_hex(num,   &r, &ERROR);
      decimal_to_hex(i,     &r2, &ERROR);
      decimal_to_hex(num*i, &r3, &ERROR);
      printf("%s * %s = %s\n", r, r2, r3);
      free(r); free(r2); free(r3);
    }
    else if(output_format == OCTAL_FORMAT)
      printf("0%o * 0%2o = 0%o\n", (int)num, (int)i, (int)(num*i));
  }
}

/*
 * Trim the given string, removing whitespace chars
 * from both ends. If KEEP_NEWLINE is TRUE, it will
 * keep newline char found at the end of the string.
 */
char *trim(char *n, char KEEP_NEWLINE)
{
  if(!n) return (char *)NULL;;
  //printf("Old '%s'(%d), ", n, strlen(n));
  int i = 0;
  int l = strlen(n);
  char NEWLINE = 0;
  if(!l) return (char *)NULL;
  if(KEEP_NEWLINE && n[l-1] == '\n') NEWLINE = 1;

  for(i = 0; i < l; i++)
  {
    if(n[i] == '\t' || n[i] == '\n' || n[i] == ' ') continue;
    else break;
  }
  n += i;
  l -= i;
  for(i = l-1; i >= 0; i--)
  {
    if(n[i] == '\t' || n[i] == '\n' || n[i] == ' ') n[i] = '\0';
    else break;
  }
  if(NEWLINE) { n[i+1] = '\n'; }
  //printf("New '%s'(%d)\n", n, strlen(n));
  return n;
}
