/* 
 *    Programmed By: Mohammed Isam Mohammed [mohammed_isam1984@yahoo.com]
 *    Copyright 2014, 2015 (c)
 * 
 *    file: numbers.c
 *    This file is part of tcalc.
 *
 *    tcalc is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    tcalc is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with tcalc.  If not, see <http://www.gnu.org/licenses/>.
 */    

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

extern int input_format, output_format;
#ifndef __FORMATS
#define __FORMATS
#define DECIMAL_FORMAT	1
#define HEX_FORMAT	2
#define BINARY_FORMAT	3
#define OCTAL_FORMAT	4
#define ALL_FORMATS	5
#endif /* __FORMATS */

char hex_chars[] =
{
  '0', '1', '2', '3', '4', '5', '6', '7', '8', '9',
  'A', 'B', 'C', 'D', 'E', 'F'
};

char is_op(char op)
{
  if(op == '+' || op == '-' ||
     op == '/' || op == '*' ||
     op == '%' || op == '^' ||
     op == '(' || op == ')' ||
     op == '\n' || op == '\t' ||
     op == '\0' || op == ' ')
    return 1;
  return 0;
}

long hex_to_decimal(char *hex, char *error)
{
  //printf("hex_to_decimal(): hex '%s'\n", hex);
  *error = 0;
  int len = strlen(hex);
  if(!hex || len < 3) goto hex_error;
  /* First make sure it is hex! */
  if(strstr(hex, "0X") != hex && strstr(hex, "0x") != hex) goto hex_error;

  long res = 0;
  int i, j = 2;
  /* Then get rid of any trailing non-hex figures */
  while(j < len)
  {
    char found = 0;
    for(i = 0; i < 16; i++)
      if(hex[j] == hex_chars[i] || hex[j] == hex_chars[i]+32)
      { found = 1; break; }
    if(found) { j++; continue; }
    break;
  }
  len = j;
  if(!is_op(hex[len])) goto hex_error;
  
  int k = (len-3)*4;
  for(j = 2; j < len; j++)
  {
    char found = 0;
    for(i = 0; i < 16; i++)
      if(hex[j] == hex_chars[i] || hex[j] == hex_chars[i]+32)
      { res += (i*(1<<k)); k -= 4; found = 1; break; }
    //printf("hex_to_decimal(): hex '%s', res '%ld', j '%d'\n", hex, res, j);
    if(!found) goto hex_error;
  }
  return res;
  
hex_error:
  printf("Error: Invalid hexadecimal input\n");
  *error = 1; return 0;
}


long binary_to_decimal(char *binary, char *error)
{
  *error = 0;
  //printf("bin_to_decimal(): binary '%s'\n", binary);
  /* Make sure it is binary! */
  if(!binary) goto binary_error;
  //if(binary[0] != 'b') goto binary_error;

  long res = 0;
  int len = strlen(binary);
  int i, j = 1;
  /* Get rid of any trailing non-numbers */
  while(j < len)
  {
    if(binary[j] != '0' && binary[j] != '1') break;
    j++;
  }
  len = j;
  if(!is_op(binary[len])) goto binary_error;

  j--;
  for(i = 0; i < len; i++)
  { res += (((binary[i] == '1') ? 1 : 0) << j); j--; }
  return res;
  
binary_error:
  printf("Error: Invalid binary input\n");
  *error = 1; return 0;
}


long octal_to_decimal(char *octal, char *error)
{
  *error = 0;
  if(!octal) goto octal_error;
  long res = 0;
  int len = strlen(octal);
  int i, j = 0;
  /* First get rid of any trailing non-numbers */
  while(j < len)
  {
    if(octal[j] < '0' || octal[j] > '7') break;
    j++;
  }
  len = j;
  if(!is_op(octal[len])) goto octal_error;

  int k = (len-1)*3;
  /* And make sure it is octal! */
  if(octal[0] != '0') goto octal_error;

  for(j = 0; j < len; j++)
  {
    if(octal[j] > '7' || octal[j] < '0') break;
    i = octal[j]-48; /* ASCII for '0' */
    res += (i*(1<<k));
    k -= 3;
  }
  //printf("octal_to_decimal(): octal '%s', res '%ld'\n", octal, res);
  return res;

octal_error:
  printf("Error: Invalid octal input\n");
  *error = 1; return 0;
}


/*long decimal_to_octal(long decimal)
{
  long res = 0, remainder = 0, quotient = 8;
  int j = 1;
  while(quotient >= 8)
  {
    quotient = decimal/8;
    remainder = decimal%8;
    res += (remainder*(10*j));
    decimal = quotient;
    j++;
  }
  res += (quotient*(10*j));
  return res;
}*/


void decimal_to_binary(long decimal, char **res, char *error)
{
  *error = 0;
  char res2[64];
  res2[63] = '\0';
  int j = 62, i;
  long quotient = 1, remainder = 0;
  while(quotient)
  {
    quotient = decimal/2;
    remainder = decimal%2;
    res2[j] = remainder ? '1' : '0';
    //printf("q=%ld, r=%ld, j=%d, res2=%c\n", quotient, remainder, j, res2[j]);
    j--;
    decimal = quotient;
  }
  //printf("%d\n", 62-j);
  *res = (char *)malloc(63-j);
  if(!*res)
  {
    printf("Error: Insufficient memory\n");
    *error = 1; return;
  }
  //char res2[62-j];
  int last = 62-j;
  for(i = 0; i < last; i++)
  {
    (*res)[i] = res2[++j];
    //printf("i=%d, res2[%d]=%c\n", i, j, res2[j]);
  }
  (*res)[i] = '\0';
  //return res2;
}


void decimal_to_hex(long decimal, char **res, char *error)
{
  *error = 0;
  char res2[64];
  sprintf(res2, "%x", (unsigned)decimal);
  int pos = 2, i = 0, j = strlen(res2);
  if(j % 2) i = 1;
  *res = (char *)malloc(j+i+2);
  if(!*res)
  {
    printf("Error: Insufficient memory\n");
    *error = 1; return;
  }
  (*res)[0] = '0';
  (*res)[1] = 'x';
  if(i) { (*res)[2] = '0'; pos++; }
  j = 0;
  for(i = 0; i < strlen(res2); i++)
    (*res)[pos++] = res2[j++];
  (*res)[pos] = '\0';
}
/*  char res2[64];
  res2[63] = '\0';
  long remainder = 0, quotient = 16;
  int j = 62, i;
  while(quotient >= 16)
  {
    quotient = decimal/16;
    remainder = decimal%16;
    res2[j] = hex_chars[remainder];
    decimal = quotient;
    j--;
  }
  res2[j] = hex_chars[quotient];
  *res = (char *)malloc(62-j);
  if(!*res)
  {
    printf("Error: Insufficient memory\n");
    return;
  }
  //char res2[62-j];
  (*res)[0] = '0';
  (*res)[1] = 'x';
  for(i = 0; i < 62-j; i++)
    (*res)[i+2] = res2[++j];
  (*res)[i+2] = '\0';
  //return res2;
}*/


void convert_numbers(char *input)
{
  if(input == NULL) return;
  long res;
  char error;
  if(input[0] == 'b' || input[0] == 'B')
  {
    res = binary_to_decimal(input+1, &error);
    if(error) return;
    if(output_format == DECIMAL_FORMAT)
      printf("%ld\n", res);
    else if(output_format == OCTAL_FORMAT)
      printf("0%o\n", (unsigned)res);
    else if(output_format == HEX_FORMAT)
    {
      char *res2;
      decimal_to_hex(res, &res2, &error);
      if(res2) { printf("%s\n", res2); free(res2); }
    }
      //printf("%x\n", (unsigned)res);
    else if(output_format == BINARY_FORMAT)
      printf("%s\n", input);
    else if(output_format == ALL_FORMATS)
    {
      printf("%ld\n", res);
      printf("0%o\n", (unsigned)res);
      char *res2;
      decimal_to_hex(res, &res2, &error);
      if(res2) { printf("%s\n", res2); free(res2); }
      //printf("%x\n", (unsigned)res);
      printf("%s\n", input);
    }
  }
  else if((input[0] == '0' && input[1] == 'X') ||
	  (input[0] == '0' && input[1] == 'x'))
  {
    res = hex_to_decimal(input, &error);
    if(error) return;
    if(output_format == DECIMAL_FORMAT)
      printf("%ld\n", res);
    else if(output_format == OCTAL_FORMAT)
      printf("0%o\n", (unsigned)res);
    else if(output_format == HEX_FORMAT)
      printf("%s\n", input);
    else if(output_format == BINARY_FORMAT)
    {
      char *res2;
      decimal_to_binary(res, &res2, &error);
      if(res2) { printf("b%s\n", res2); free(res2); }
    }
    else if(output_format == ALL_FORMATS)
    {
      printf("%ld\n", res);
      printf("0%o\n", (unsigned)res);
      printf("%s\n", input);
      char *res2;
      decimal_to_binary(res, &res2, &error);
      if(res2) { printf("b%s\n", res2); free(res2); }
    }
  }
  else if(input[0] == '0')
  {
    res = octal_to_decimal(input, &error);
    if(error) return;
    if(output_format == DECIMAL_FORMAT)
      printf("%ld\n", res);
    else if(output_format == OCTAL_FORMAT)
      printf("%s\n", input);
    else if(output_format == HEX_FORMAT)
    {
      char *res2;
      decimal_to_hex(res, &res2, &error);
      if(res2) { printf("%s\n", res2); free(res2); }
    }
      //printf("%x\n", (unsigned)res);
    else if(output_format == BINARY_FORMAT)
    {
      char *res2;
      decimal_to_binary(res, &res2, &error);
      if(res2) { printf("b%s\n", res2); free(res2); }
    }
    else if(output_format == ALL_FORMATS)
    {
      printf("%ld\n", res);
      printf("%s\n", input);
      char *res2;
      decimal_to_hex(res, &res2, &error);
      if(res2) { printf("%s\n", res2); free(res2); }
      //printf("%x\n", (unsigned)res);
      decimal_to_binary(res, &res2, &error);
      if(res2) { printf("b%s\n", res2); free(res2); }
    }
  }
  else
  {
    res = atol(input);
    if(output_format == DECIMAL_FORMAT)
      printf("%ld\n", res);
    else if(output_format == OCTAL_FORMAT)
      printf("0%o\n", (unsigned)res);
    else if(output_format == HEX_FORMAT)
    {
      char *res2;
      decimal_to_hex(res, &res2, &error);
      if(res2) { printf("%s\n", res2); free(res2); }
    }
      //printf("0x%02x\n", (unsigned)res);
    else if(output_format == BINARY_FORMAT)
    {
      char *res2;
      decimal_to_binary(res, &res2, &error);
      if(res2) { printf("b%s\n", res2); free(res2); }
    }
    else if(output_format == ALL_FORMATS)
    {
      printf("%ld\n", res);
      printf("0%o\n", (unsigned)res);
      char *res2;
      decimal_to_hex(res, &res2, &error);
      if(res2) { printf("%s\n", res2); free(res2); }
      //printf("%x\n", (unsigned)res);
      decimal_to_binary(res, &res2, &error);
      if(res2) { printf("b%s\n", res2); free(res2); }
    }
  }
}